#!/usr/bin/env python
#######################
#
#    This program generates heatmaps of numeric data. Three data columns of data
#    are required: the x axis labels, y axis labels and an intensity value. An optional
#    grouping value can also be included. The input file uses a zero-based numbering
#    system, so the first column is 0, the second is 1, etc.
#    The command line syntax is:
#
#        python heatmap_cmd.py [options]
#
#		 options: 
#        --in_dir  		Name of input directory (followed by /)
#        --in_file 		Name of input file
#        --group_col    Column for heatmap grouping
#        --x_col        Column for x values
#        --y_col        Column for y values
#        --value_col    Column of heatmap intensity values
#        --out_dir      Name of output directory (ignore this option to make it same as input directory)
#        --bg_dir       Name of background directory (followed by /)
#        --bg_group     ID of common background group to use
#
##################

### Import modules
import matplotlib.pyplot as plt
from pylab import *
import sys
import numpy as np
import math
import csv
import datetime
import os
from matplotlib import rcParams
from operator import add
import matplotlib.font_manager
import argparse

print sys.argv

parser = argparse.ArgumentParser(description='Take inputs for creating heatmaps')
parser.add_argument('--in_dir',   type=str, help='Name of input directory (followed by /)')
parser.add_argument('--in_file',  type=str, help='Name of inptu file')
parser.add_argument('--group_col', type=int,help='Column ID for heatmap grouping')
parser.add_argument('--x_col',   type=int, help='Column ID for x values')
parser.add_argument('--y_col',type=int, help='Column ID for y values')
parser.add_argument('--value_col',    type=int, help='Column of heatmap intensity values')
parser.add_argument('--count_col',  type=int, help='Column ID for count number of each intensity')
parser.add_argument('--out_dir',  type=str, help='Name of output directory (ignore this value make it same as input directory')
parser.add_argument('--bg_dir',  type=str, help='Name of background directory (followed by /)')
parser.add_argument('--bg_group',  type=str, help='Name of background group')

def main():

    # Get the command line arguments
    args = get_args()

    # Make the input and heatmap settings dictionaries
    input_settings, heatmap_settings = configure_settings(args)

    # Open the main heatmap data file
    data = import_data_from_file(input_settings['input_dir']+input_settings['input_file'])

    # Arrange the main heatmap data in a data dictionary
    data_dict = parse_data_from_lines(input_settings, data)

    # Create sorted list of groups
    group_list = sorted(list(data_dict.keys()))

    # Create output directory (if necessary) and prepare output file
    if not os.path.exists(input_settings['output_dir']):
        os.mkdir(input_settings['output_dir'])
    ofile  = open((input_settings['output_dir']+input_settings['outputfile']), "wb")
    writer = csv.writer(ofile, delimiter='\t')
    if heatmap_settings['remove_low_counts'] is True:
        headers = ['Group','X','Y','Value', 'Counts']
    else:
        headers = ['Group','X','Y','Value']

    # Write main data file to output
    write_to_output(data_dict,group_list,writer,headers,[str(input_settings['input_file']+' - Original')])

    # If background must be removed, process background file, and add background data and subtracted main file to output
    if input_settings['subtract_background'] is True:
        bg_data = import_data_from_file(input_settings['input_dir']+input_settings['bg_inputfile'])
        bg_data_dict = parse_data_from_lines(input_settings, bg_data)    
        bg_group_list = sorted(list(bg_data_dict.keys()))
        data_dict = subtract_background(input_settings['background_group'],data_dict, group_list,bg_data_dict, bg_group_list)
        write_to_output(bg_data_dict,bg_group_list,writer,headers,[input_settings['bg_inputfile']+' - Background Data'])
        write_to_output(data_dict,group_list,writer,headers,[input_settings['input_file']+' - After Subtraction'])
    print "group list: ", group_list
    print "groups to process", str(heatmap_settings['groups_to_process'])
    # Initialize the heatmap figure
    fig = init_heatmaps(heatmap_settings)

    # For each group,organize data into correct format and make heatmaps
    for i in group_list:
        if (str(i) in str(heatmap_settings['groups_to_process'])) or (heatmap_settings['use_all_groups'] is True) :
            label_keys = [[],[]]
            #print "input group:", data_dict[i]
            group, data_array, label_keys = organize_data(input_settings, heatmap_settings,data_dict[i],label_keys,i,group_list.index(i)+1)
            #print "output group:", group
            data_dict[str(i)] = group
            make_heatmap(fig,heatmap_settings,data_array,label_keys,i,group_list.index(i)+1)
    write_to_output(data_dict,heatmap_settings['groups_to_process'], writer,headers,[input_settings['input_file']+' - Insignificant Data Removed'])

    # Close the output file
    ofile.close()

    # Save heatmap figure to a file
    save_to_file(input_settings, heatmap_settings)

    print fig.get_size_inches()


    # Display the heatmap
    if heatmap_settings['display_heatmaps'] is True:
        plt.show()

    # Clear the image buffer before exiting
    plt.cla()
    plt.clf()
    plt.close()
    sys.exit(0)

def get_args():
    '''
    Return the argparse arguments as a dictionary

    '''
    args = vars(parser.parse_args())
    return args

def init_heatmaps(heatmap_settings):

    # This function is used to initialize the heatmaps and create a figure with
    # the correct spacing.

    fig = plt.figure(figsize=[heatmap_settings['plot_width'], heatmap_settings['plot_height']])
    fig.subplots_adjust(bottom=heatmap_settings['subplot_adjustments']['bottom'],
                        top=heatmap_settings['subplot_adjustments']['top'],
                        left=heatmap_settings['subplot_adjustments']['left'],
                        right=heatmap_settings['subplot_adjustments']['right'],
                        wspace=heatmap_settings['subplot_adjustments']['wspace']
                        )
    return fig

def subtract_background(bg_group,data_dict, group_list,bg_data_dict, bg_group_list):

    # This function subtracts the values of one dataset from another. Only values
    # that have an identical group, x and y value will be subtracted. The other
    # values from both datasets will be ignored. It returns the updated dataset.

    for i in bg_group_list:
        if i in group_list:
            for j in range(len(bg_data_dict[i][0])):
                if bg_group is None:
                    loc = (bg_data_dict[i][0][j], bg_data_dict[i][1][j])
                    print "Subtracting each group"
                    z = i
                else:
                    z = bg_group
                    loc = (bg_data_dict[z][0][j], bg_data_dict[z][1][j])
                    print "Subtracting only using data from group: ", bg_group
                for k in range(len(data_dict[i][0])):
                    if data_dict[i][0][k] is loc[0] and data_dict[i][1][k] is loc[1]:
                        data_dict[i][2][k] = float(data_dict[i][2][k]) - float(bg_data_dict[z][2][j])

    return data_dict


def configure_settings(args):

    # Initializes the file input and matplotlib settings. The heatmap_settings
    # can be changed, but the input settings should not be. The function returns
    # the input and heatmap settings.

    ########## Do not change these!
    input_settings = {
        'input_dir':args['in_dir'],
        'input_file':args['in_file'],
        'group_col':args['group_col'],
        'x_col':args['x_col'],
        'y_col':args['y_col'],
        'value_col':args['value_col'],
        'background_group':args['bg_group']
        }


    # Set the output directory
    if args['out_dir'] is None:
        input_settings['output_dir'] = input_settings['input_dir']+'/'
    else:
        input_settings['output_dir'] = args['out_dir']

    # Set the background data
    if args['bg_dir'] is None:
        input_settings['subtract_background'] = False
    else:
        input_settings['subtract_background'] = True
        input_settings['bg_inputfile'] = args['bg_dir'] + args['in_file']

    input_settings['outputfile'] = input_settings['input_file'].split('.')[0]+'_HeatMapOutput.txt'
    input_settings['outputimage'] = input_settings['input_file'].split('.')[0]+'_HeatMap'
    ###########

    ########### These can be changed as needed to customize the heatmaps
    heatmap_settings = {
        'show_title':False,
        'show_xlabel':False,
        'show_ylabel':False,
        'show_colorbar':False,
        'x_axis_label':'Category X',
        'y_axis_label':'Category Y',
        'file_label':input_settings['input_file'],
        'show_file_label':False,
        'alpha_labels':True,
        'label_font':{'family':'sans serif','color':'black','weight':'normal','size':8},
        'title_font':{'family':'sans serif','color':'black','weight':'normal','size':10},
        'tick_label_size':10,
        'tick_label_color':'black',
        'show_x_tick_labels':False,
        'show_y_tick_labels':False,
        'show_ticks':{'left':False, 'right':False,'top':False,'bottom':False},
        'border_width':0,
        'use_all_groups':False,
        'groups_to_process':['15','16','17','18','19','20','21','22','23','24','25','26'],
        'number_of_subplots':12,
        'remove_insig_data':False,
        'remove_low_counts':True,
        'sig_value':.15,
        'sig_count':50,
        'export_svg':True,
        'export_png':False,
        'export_tif':True,
        'cmap':'PiYG',
        'norm_settings':mpl.colors.Normalize(vmin=-0.5, vmax=0.5),
        'plot_height':.048,
        'plot_width':.6,
        'dpi':600,
        'heatmap_edgecolor':'none',
        'savefig_facecolor':'white',
        'subplot_adjustments':{'bottom':0.02,'top':0.98,'left':0.02,'right':0.98,'wspace':0.12},
        'display_heatmaps':False,
        'show_spine':{'bottom':True,'top':True,'left':True,'right':True},
        'spine_width':{'bottom':0.1,'top':0.1,'left':0.1,'right':0.1}
        }
    #############

    # Set the count column
    if args['count_col'] is None:
        print "count col is none"
        input_settings['count_col'] = None
        heatmap_settings['remove_low_counts'] = False
    else:
        input_settings['count_col'] = args['count_col']

    return (input_settings, heatmap_settings)


def import_data_from_file(filename):

    # This function opens a file from the disk. It returns the raw data.

    try:
        return open(filename)
    except IOError:
        print "Unable to open file: %s" %(filename)


def parse_data_from_lines(input_settings,datafile):

    # This functions parses the raw data into a format that can be easily
    # used, based on the command line inputs. It returns a data dictionary

    data_dict = dict()
    try:
        for line in datafile:
            split_line = line.strip('\n').split('\t')
            y_label = str(split_line[input_settings['y_col']])
            x_label = str(split_line[input_settings['x_col']])
            value = split_line[input_settings['value_col']]
            if input_settings['count_col'] != None:
                print 'getting counts column'
                counts = int(split_line[input_settings['count_col']])
            else:
                counts = ''
            if input_settings['group_col'] is not None:
                group = str(split_line[input_settings['group_col']])
            else:
                group = 0
            data_dict = arrange_grid(data_dict,group,x_label,y_label,value,counts)
        return data_dict
    except StandardError:
        print "There was an error parsing the datafile"


def arrange_grid(data_dict,group, X, Y, value,counts):

    # This function is called by the 'parse_data_from_lines' function only. It
    # is used to organize the data in lists. It returns the data dictionary.

    if group in data_dict.keys():
        data_dict[group][0].append(X)
        data_dict[group][1].append(Y)
        data_dict[group][2].append(value)
        data_dict[group][3].append(counts)
    else:
        data_dict[group] = ([],[],[],[])
        data_dict[group][0].append(X)
        data_dict[group][1].append(Y)
        data_dict[group][2].append(value)
        data_dict[group][3].append(counts)
    return data_dict


def organize_data(input_settings, heatmap_settings,group, label_keys,group_label,subplot_number):

    # This function creates the data array that is sent to the Matplotlib function
    # to make the heatmap. It also identifies and orders the labels for each
    # axis. It returns the data array and the label keys.
    #print "this is the data for the group: ", group
    if heatmap_settings['alpha_labels'] is True:
        unique_x = sorted((list(set(group[0]))),key=str)
        unique_y = sorted((list(set(group[1]))),key=str)
    else:
        unique_x = sorted((list(set(group[0]))),key=int)
        unique_y = sorted((list(set(group[1]))),key=int)

    for i in range(len(unique_x)):
        label_keys[0].append((unique_x[i], i))
    for i in range(len(unique_y)):
        label_keys[1].append((unique_y[i], i))

    if heatmap_settings['remove_low_counts'] == True or heatmap_settings['remove_insig_data'] == True:
        group = check_significance(input_settings['count_col'], heatmap_settings['sig_count'], heatmap_settings['remove_insig_data'], heatmap_settings['sig_value'],group, unique_x)

        #print 'label_keys 0', label_keys[0]
        #print 'label_keys 1', label_keys[1]
    data_array = np.zeros(((int(len(unique_y))),(int(len(unique_x)))))

    for i in range(len(group[0])):
        data_array[[item for item in label_keys[1] if item[0] == (group[1][i])][0][1]][[item for item in label_keys[0] if item[0] == (group[0][i])][0][1]] = (group[2][i])

    #if heatmap_settings['remove_insig_data'] is True:
    #    data_array = remove_insignificant_data(heatmap_settings, data_array, len(unique_x), len(unique_y))
    #if heatmap_settings['remove_low_counts'] is True:
    #    data_array = remove_low_counts(input_settings['count_col'], heatmap_settings['sig_count'], data_array)

    return (group, data_array, label_keys)

def check_significance(count_col,sig_count, check_sig, sig_value, group, x_labels):
    # This function removes any data where there were too few N to make a meaningful interpretation
    #print "this is the count col:", count_col
    #print "this is the sig count:", sig_count
    #print "this is the group:", group
    #print "this is the x_label_keys:", x_labels
    

    if check_sig == False:
        sig_value = 0;
    for x in x_labels:
        count = 0
        values = []
        for i,j in enumerate(group[0]):
            if x == j:
                count += group[3][i]
                values.append(group[2][i])
        if count < sig_count or ((max(values)-min(values) < sig_value)):
            for i,j in enumerate(group[0]):
                if j == x:
                    group[2][i] = 0
        #print "for this key, the total is: ", count
        #print "for this group, the values are: ", values, "range: ", (max(values)-min(values))
        #print "this is the new group:", group
    return group


def make_heatmap(fig, heatmap_settings,data_array,label_keys,group_label,subplot_number):

    # This function calls the Matplotlib function pcolormesh to create the
    # heatmaps. It also performs the formatting for the image based on the
    # settings file

    ax = plt.subplot(1,heatmap_settings['number_of_subplots'],subplot_number)
    #plt.pcolormesh(data_array,cmap=heatmap_settings['cmap'],norm=heatmap_settings['norm_settings'],lw=heatmap_settings['border_width'],edgecolor=heatmap_settings['heatmap_edgecolor'])
    plt.pcolor(data_array,cmap=heatmap_settings['cmap'],norm=heatmap_settings['norm_settings'],lw=heatmap_settings['border_width'],edgecolor=heatmap_settings['heatmap_edgecolor'])

    x_labels = []
    y_labels = []

    for i in label_keys[0]:
        x_labels.append(i[0])
    for i in label_keys[1]:
        y_labels.append(i[0])

    ax.set_yticks(np.arange(len(data_array))+0.5)
    ax.set_xticks(np.arange(len(data_array[0]))+0.5)

    # Set or hide tick labels
    if heatmap_settings['show_x_tick_labels'] is False:
        ax.set_xticklabels([], minor=False)
    else:
        ax.set_xticklabels(x_labels, minor=False)

    if heatmap_settings['show_y_tick_labels'] is False:
        ax.set_yticklabels([], minor=False)
    else:
        ax.set_yticklabels(y_labels, minor=False)

    # Set visibility of spines
    ax.spines['bottom'].set_visible(heatmap_settings['show_spine']['bottom'])
    ax.spines['top'].set_visible(heatmap_settings['show_spine']['top'])
    ax.spines['left'].set_visible(heatmap_settings['show_spine']['left'])
    ax.spines['right'].set_visible(heatmap_settings['show_spine']['right'])
    ax.spines['bottom'].set_linewidth(heatmap_settings['spine_width']['bottom'])
    ax.spines['top'].set_linewidth(heatmap_settings['spine_width']['top'])
    ax.spines['left'].set_linewidth(heatmap_settings['spine_width']['left'])
    ax.spines['right'].set_linewidth(heatmap_settings['spine_width']['right'])
    ax.tick_params(axis='x', which='major', labelsize=heatmap_settings['tick_label_size'], labelcolor=heatmap_settings['tick_label_color'],direction='out',top='on',bottom='off',labeltop=heatmap_settings['show_x_tick_labels'],pad=-2)
    ax.tick_params(axis='y', which='major', labelsize=heatmap_settings['tick_label_size'], labelcolor=heatmap_settings['tick_label_color'],direction='out',left='on',right='off',labelleft=heatmap_settings['show_y_tick_labels'])

    if subplot_number > 1:
        ax.tick_params(axis='y', left='off',labelleft=False)
    else:
        if heatmap_settings['show_file_label'] is True:
            plt.ylabel(heatmap_settings['file_label'],fontsize=5,labelpad=20)

    # Set ticks
    ax.tick_params(left=heatmap_settings['show_ticks']['left'],
        right=heatmap_settings['show_ticks']['right'],
        top=heatmap_settings['show_ticks']['top'],
        bottom=heatmap_settings['show_ticks']['bottom'])
    ax.invert_yaxis()

    return

def write_to_output(data_dict, group_list,writer,headers,data_label):

    # This function writes the raw data to the output file in tab-delimited
    # format

    writer.writerow(data_label)
    for ID in group_list:
        writer.writerow(headers)
        print "ID ", ID
        for i,j in enumerate(data_dict[str(ID)][0]):
            print "j ", j
            print "i ", i
            if data_dict[ID][3] is not None:
                writer.writerow((ID, data_dict[ID][0][i], data_dict[ID][1][i], data_dict[ID][2][i], data_dict[ID][3][i]))
            else:
                writer.writerow((ID, data_dict[ID][0][i], data_dict[ID][1][i], data_dict[ID][2][i]))
        writer.writerow('')
    writer.writerow('')
    writer.writerow('')


def save_to_file(input_settings,heatmap_settings):

    # This function saves the heatmaps to image files

    if heatmap_settings['export_svg'] is True:
        plt.savefig(input_settings['output_dir']+input_settings['outputimage']+'.svg', format='svg', dpi=heatmap_settings['dpi'], facecolor=heatmap_settings['savefig_facecolor'])
        #print "printing svg"
    if heatmap_settings['export_png'] is True:
        plt.savefig(input_settings['output_dir']+input_settings['outputimage']+'.png', format='png', dpi=heatmap_settings['dpi'], facecolor=heatmap_settings['savefig_facecolor'])
        #print "printing png"
    if heatmap_settings['export_tif'] is True:
        plt.savefig(input_settings['output_dir']+input_settings['outputimage']+'.tif', format='tif', dpi=heatmap_settings['dpi'], facecolor=heatmap_settings['savefig_facecolor'])
        #print "printing tif"
    return


if __name__ == "__main__":
    sys.exit(main())


